@@ -10,15 +10,17 @@ mode and Spyder
 """
 
 # Standard library imports
+import atexit
 import pdb
 import re
 
 # Third-party imports
 from IPython.core.history import HistoryManager
 from IPython.core.inputtransformer2 import TransformerManager
-from IPython.lib.lexers import IPythonLexer, IPython3Lexer
-from pygments.lexer import bygroups
-from pygments.token import Keyword, Operator, Text
+from ipython_pygments_lexers import IPython3Lexer
+from pygments.lexer import bygroups, using
+from pygments.lexers import Python3Lexer
+from pygments.token import Keyword, Operator
 from pygments.util import ClassNotFound
 from qtconsole.rich_jupyter_widget import RichJupyterWidget
 from qtpy.QtCore import QEvent
@@ -32,14 +34,11 @@ from spyder.config.base import get_conf_path
 class SpyderIPy3Lexer(IPython3Lexer):
     # Detect !cmd command and highlight them
     tokens = IPython3Lexer.tokens
-    tokens['root'].insert(
-        0, (r'(!)(\w+)(.*\n)', bygroups(Operator, Keyword, Text)))
-
-
-class SpyderIPy2Lexer(IPythonLexer):
-    tokens = IPython3Lexer.tokens
-    tokens['root'].insert(
-        0, (r'(!)(\w+)(.*\n)', bygroups(Operator, Keyword, Text)))
+    spyder_tokens = [
+        (r'(!)(\w+)(.*\n)', bygroups(Operator, Keyword, using(Python3Lexer))),
+        (r'(%)(\w+)(.*\n)', bygroups(Operator, Keyword, using(Python3Lexer))),
+    ]
+    tokens['root'] = spyder_tokens + tokens['root']
 
 
 class PdbHistory(HistoryManager):
@@ -64,7 +63,7 @@ class DebuggingHistoryWidget(RichJupyterWidget):
     def __init__(self, *args, **kwargs):
         # History
         self._pdb_history_input_number = 0  # Input number for current session
-        self._saved_pdb_history_input_number = []  # for recursive debugging
+        self._saved_pdb_history_input_number = {}  # for recursive debugging
 
         # Catch any exception that prevents to create or access the history
         # file to avoid errors.
@@ -190,13 +189,15 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
 
     def __init__(self, *args, **kwargs):
         # Communication state
-        self._pdb_in_loop = 0  # NUmber of debbuging loop we are in
+        self._pdb_recursion_level = 0  # Number of debbuging loop we are in
         self._pdb_input_ready = False  # Can we send a command now
         self._waiting_pdb_input = False  # Are we waiting on the user
         # Other state
-        self._pdb_prompt = (None, None)  # prompt, password
+        self._pdb_prompt = None  # prompt
+        self._pdb_prompt_input = False  # wether pdb waits for input or comm
         self._pdb_last_cmd = ''  # last command sent to pdb
         self._pdb_frame_loc = (None, None)  # fname, lineno
+        self._pdb_take_focus = True  # Focus to shell after command execution
         # Command queue
         self._pdb_input_queue = []  # List of (code, hidden, echo_stack_entry)
         # Temporary flags
@@ -211,15 +212,27 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
             r'[ \t]*\(*IPdb \[\d+\]\)*: |' +
             r'[ \t]*In \[\d+\]: |[ \t]*\ \ \ \.\.\.+: )')
 
+        # Reset debug state when debugging is done
+        self.sig_prompt_ready.connect(self.reset_debug_state)
+
     # --- Public API --------------------------------------------------
 
     def shutdown(self):
         """
         Close the save thread and database file.
         """
+        try:
+            # Make sure the database will not be called after closing
+            self.sig_prompt_ready.disconnect(self.reset_debug_state)
+        except (TypeError, RuntimeError):
+            # Already disconnected
+            pass
+
         if self._pdb_history_file is not None:
             try:
                 self._pdb_history_file.save_thread.stop()
+                # Now that it was called, no need to call it at exit
+                atexit.unregister(self._pdb_history_file.save_thread.stop)
             except AttributeError:
                 pass
 
@@ -230,36 +243,39 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
 
     # --- Comm API --------------------------------------------------
 
-    def set_debug_state(self, is_debugging):
+    def set_debug_state(self, recursion_level):
         """Update the debug state."""
-        if is_debugging:
+        if recursion_level == self._pdb_recursion_level:
+            # Nothing to change
+            return
+
+        if recursion_level > self._pdb_recursion_level:
             # Start debugging
-            if self._pdb_in_loop > 0:
-                # Recursive debugging
-                self._saved_pdb_history_input_number.append(
-                    self._pdb_history_input_number)
+            if self._pdb_recursion_level > 0:
+                # Recursive debugging, save state
+                self._saved_pdb_history_input_number[
+                    self._pdb_recursion_level] = self._pdb_history_input_number
                 self.end_history_session()
             self.new_history_session()
-            self._pdb_in_loop += 1
-        elif self._pdb_in_loop > 0:
+        elif recursion_level < self._pdb_recursion_level:
             # Stop debugging
-            self._pdb_in_loop -= 1
             self.end_history_session()
-            if self._pdb_in_loop > 0:
-                # Still debugging
+            if recursion_level > 0:
+                # Still debugging, restore state
                 self.new_history_session()
                 self._pdb_history_input_number = (
-                    self._saved_pdb_history_input_number.pop())
+                    self._saved_pdb_history_input_number.pop(
+                        recursion_level, 0))
 
         # If debugging starts or stops, clear the input queue.
+        self._pdb_recursion_level = recursion_level
         self._pdb_input_queue = []
         self._pdb_frame_loc = (None, None)
 
     def _pdb_cmd_prefix(self):
         """Return the command prefix"""
         prefix = ''
-        if (self.spyder_kernel_comm.is_open() and
-                self.is_pdb_using_exclamantion_mark()):
+        if self.spyder_kernel_ready and self.is_pdb_using_exclamantion_mark():
             prefix = '!'
         return prefix
 
@@ -267,23 +283,35 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
         """
         Execute a pdb command
         """
+        self._pdb_take_focus = False
         self.pdb_execute(
             self._pdb_cmd_prefix() + command, hidden=False,
             echo_stack_entry=False, add_history=False)
 
     def _handle_input_request(self, msg):
         """Process an input request."""
-        if (not self.spyder_kernel_comm.is_open() and
-                msg['content']['prompt'] == "ipdb> "):
+        if not self.is_spyder_kernel and "ipdb>" in msg['content']['prompt']:
             # Check if we can guess a path from the shell content:
             self._flush_pending_stream()
             cursor = self._get_end_cursor()
             cursor.setPosition(self._prompt_pos, QTextCursor.KeepAnchor)
             text = cursor.selection().toPlainText()
             match = re.search(r"> (.*\.py)\((\d+)\)", text)
+            state = None
+
             if match:
                 fname, lineno = match.groups()
-                self.sig_pdb_step.emit(fname, int(lineno))
+                state = {
+                    'step': {
+                        'fname': fname,
+                        'lineno': int(lineno)
+                    }
+                }
+
+            prompt = msg['content']['prompt']
+            password = msg['content']['password']
+            self.pdb_input(prompt, password, state, from_input=True)
+            return
         return super(DebuggingWidget, self)._handle_input_request(msg)
 
     def pdb_execute(self, line, hidden=False, echo_stack_entry=True,
@@ -305,14 +333,6 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
         add_history: bool
             If not hidden, wether the line should be added to history
         """
-        # Send line to input if no comm
-        if not self.spyder_kernel_comm.is_open():
-            if not hidden:
-                self._append_plain_text(line + '\n')
-            self._finalize_input_request()
-            self.kernel_client.input(line)
-            return
-
         if not self.is_debugging():
             return
 
@@ -354,65 +374,30 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
 
             # Emit executing
             self.executing.emit(line)
-            self.sig_pdb_state_changed.emit(
-                False, self.get_pdb_last_step())
+            self.sig_pdb_state_changed.emit(False)
 
         if self._pdb_input_ready:
             # Print the string to the console
             self._pdb_input_ready = False
-            return self.call_kernel(interrupt=True).pdb_input_reply(
-                line, echo_stack_entry=echo_stack_entry)
+            self.pdb_input_reply(line, echo_stack_entry)
+            return
 
         self._pdb_input_queue.append(
             (line, hidden, echo_stack_entry, add_history))
 
-    def get_pdb_settings(self):
-        """Get pdb settings"""
-        return {
-            "breakpoints": self.get_conf(
-                'breakpoints', default={}, section='run'),
-            "pdb_ignore_lib": self.get_conf('pdb_ignore_lib'),
-            "pdb_execute_events": self.get_conf('pdb_execute_events'),
-            "pdb_use_exclamation_mark": self.is_pdb_using_exclamantion_mark(),
-            "pdb_stop_first_line": self.get_conf('pdb_stop_first_line'),
-        }
+    def reset_debug_state(self):
+        """Reset debug state if the debugger crashed."""
+        self.set_debug_state(0)
 
     # --- To Sort --------------------------------------------------
     def stop_debugging(self):
         """Stop debugging."""
-        if (self.spyder_kernel_comm.is_open() and
-                not self.is_waiting_pdb_input()):
+        if self.spyder_kernel_ready and not self.is_waiting_pdb_input():
             self.interrupt_kernel()
         self.pdb_execute_command("exit")
 
-    def set_spyder_breakpoints(self):
-        """Set Spyder breakpoints into a debugging session"""
-        self.call_kernel(interrupt=True).set_breakpoints(
-            self.get_conf('breakpoints', default={}, section='run'))
-
-    def set_pdb_ignore_lib(self, pdb_ignore_lib):
-        """Set pdb_ignore_lib into a debugging session"""
-        self.call_kernel(interrupt=True).set_pdb_ignore_lib(
-            pdb_ignore_lib)
-
-    def set_pdb_execute_events(self, pdb_execute_events):
-        """Set pdb_execute_events into a debugging session"""
-        self.call_kernel(interrupt=True).set_pdb_execute_events(
-            pdb_execute_events)
-
-    def set_pdb_use_exclamation_mark(self, pdb_use_exclamation_mark):
-        """Set pdb_use_exclamation_mark into a debugging session"""
-        self.call_kernel(interrupt=True).set_pdb_use_exclamation_mark(
-            pdb_use_exclamation_mark)
-
     def is_pdb_using_exclamantion_mark(self):
-        return self.get_conf('pdb_use_exclamation_mark')
-
-    def do_where(self):
-        """Where was called, go to the current location."""
-        fname, lineno = self._pdb_frame_loc
-        if fname:
-            self.sig_pdb_step.emit(fname, lineno)
+        return self.get_conf('pdb_use_exclamation_mark', section='debugger')
 
     def refresh_from_pdb(self, pdb_state):
         """
@@ -421,9 +406,10 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
 
         See publish_pdb_state and notify_spyder in spyder_kernels
         """
-        if 'step' in pdb_state and 'fname' in pdb_state['step']:
-            fname = pdb_state['step']['fname']
-            lineno = pdb_state['step']['lineno']
+        pdb_step = pdb_state.pop('step', None)
+        if pdb_step and 'fname' in pdb_step:
+            fname = pdb_step['fname']
+            lineno = pdb_step['lineno']
 
             last_pdb_loc = self._pdb_frame_loc
             self._pdb_frame_loc = (fname, lineno)
@@ -432,16 +418,21 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
             if (fname, lineno) != last_pdb_loc:
                 self.sig_pdb_step.emit(fname, lineno)
 
-        if 'namespace_view' in pdb_state:
-            self.set_namespace_view(pdb_state['namespace_view'])
+        if "do_where" in pdb_state:
+            fname, lineno = self._pdb_frame_loc
+            if fname:
+                self.sig_pdb_step.emit(fname, lineno)
 
-        if 'var_properties' in pdb_state:
-            self.set_var_properties(pdb_state['var_properties'])
+        pdb_stack = pdb_state.pop('stack', None)
+        if pdb_stack:
+            pdb_stack, pdb_index = pdb_stack
+            self.sig_pdb_stack.emit(pdb_stack, pdb_index)
 
-    def set_pdb_state(self, pdb_state):
-        """Set current pdb state."""
-        if pdb_state is not None and isinstance(pdb_state, dict):
-            self.refresh_from_pdb(pdb_state)
+        request_pdb_input =  pdb_state.pop('request_pdb_input', None)
+        if request_pdb_input:
+            self.pdb_execute(request_pdb_input)
+
+        self.update_state(pdb_state)
 
     def show_pdb_output(self, text):
         """Show Pdb output."""
@@ -460,19 +451,19 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
 
     def get_pdb_last_step(self):
         """Get last pdb step retrieved from a Pdb session."""
-        fname, lineno = self._pdb_frame_loc
-        if fname is None:
-            return {}
-        return {'fname': fname,
-                'lineno': lineno}
+        return self._pdb_frame_loc
 
     def is_debugging(self):
         """Check if we are debugging."""
-        return self._pdb_in_loop > 0
+        return self._pdb_recursion_level > 0
+
+    def debugging_depth(self):
+        """Debugging depth"""
+        return self._pdb_recursion_level
 
     def is_waiting_pdb_input(self):
         """Check if we are waiting a pdb input."""
-        # If the comm is not open, self._pdb_in_loop can not be set
+        # If the comm is not open, self._pdb_recursion_level can not be set
         return self.is_debugging() and self._waiting_pdb_input
 
     # ---- Public API (overrode by us) ----------------------------
@@ -484,16 +475,24 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
         super(DebuggingWidget, self).reset(clear)
         # Make sure the prompt is printed
         if clear and self.is_waiting_pdb_input():
-            prompt, password = self._pdb_prompt
-            self.kernel_client.iopub_channel.flush()
+            prompt = self._pdb_prompt
+
+            try:
+                # This is necessary to avoid an error when the iopub channel is
+                # closed.
+                # See jupyter/qtconsole#574
+                if not self.kernel_client.iopub_channel.closed():
+                    self.kernel_client.iopub_channel.flush()
+            except AttributeError:
+                self.kernel_client.iopub_channel.flush()
+
             self._reading = False
-            self._readline(prompt=prompt, callback=self._pdb_readline_callback,
-                           password=password)
+            self._readline(prompt=prompt, callback=self.pdb_execute)
 
     # --- Private API --------------------------------------------------
     def _current_prompt(self):
         prompt = "IPdb [{}]".format(self._pdb_history_input_number + 1)
-        for i in range(self._pdb_in_loop - 1):
+        for i in range(self._pdb_recursion_level - 1):
             # Add recursive debugger prompt
             prompt = "({})".format(prompt)
         return prompt + ": "
@@ -501,7 +500,7 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
     def _current_out_prompt(self):
         """Get current out prompt."""
         prompt = "Out\u00A0\u00A0[{}]".format(self._pdb_history_input_number)
-        for i in range(self._pdb_in_loop - 1):
+        for i in range(self._pdb_recursion_level - 1):
             # Add recursive debugger prompt
             prompt = "({})".format(prompt)
         return prompt + ": "
@@ -515,8 +514,6 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
             # add custom lexer
             if pygments_lexer == 'ipython3':
                 lexer = SpyderIPy3Lexer()
-            elif pygments_lexer == 'ipython2':
-                lexer = SpyderIPy2Lexer()
             else:
                 return
             self._highlighter._lexer = lexer
@@ -529,9 +526,9 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
         original_complete = client.complete
 
         def complete(code, cursor_pos=None):
-            if self.is_waiting_pdb_input() and client.comm_channel:
+            if self.is_waiting_pdb_input():
                 shell_channel = client.shell_channel
-                client._shell_channel = client.comm_channel
+                client._shell_channel = client.control_channel
                 try:
                     return original_complete(code, cursor_pos)
                 finally:
@@ -541,15 +538,13 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
 
         client.complete = complete
 
-    def _update_pdb_prompt(self, prompt, password=None):
+    def _update_pdb_prompt(self, prompt):
         """Update the prompt that is recognised as a pdb prompt."""
-        if prompt == self._pdb_prompt[0]:
+        if prompt == self._pdb_prompt:
             # Nothing to do
             return
 
-        if password is None:
-            password = self._pdb_prompt[1]
-        self._pdb_prompt = (prompt, password)
+        self._pdb_prompt = prompt
 
         # Update continuation prompt to reflect (possibly) new prompt length.
         self._set_continuation_prompt(
@@ -604,16 +599,18 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
         return super(DebuggingWidget, self).execute(
             source, hidden, interactive)
 
-    def _pdb_readline_callback(self, line):
-        """Callback used when the user inputs text in pdb."""
-        self.pdb_execute(line)
-
-    def pdb_input(self, prompt, password=None):
+    def pdb_input(self, prompt, password=None, state=None, from_input=False):
         """Get input for a command."""
 
+        self.set_debug_state(1 + prompt.count("("))
+
+        if state is not None and isinstance(state, dict):
+            self.refresh_from_pdb(state)
+
         # Replace with numbered prompt
         prompt = self._current_prompt()
-        self._update_pdb_prompt(prompt, password)
+        self._update_pdb_prompt(prompt)
+        self._pdb_prompt_input = from_input
 
         # The prompt should be printed unless:
         # 1. The prompt is already printed (self._reading is True)
@@ -625,16 +622,25 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
         if print_prompt:
             # Make sure that all output from the SUB channel has been processed
             # before writing a new prompt.
-            self.kernel_client.iopub_channel.flush()
+            try:
+                # This is necessary to avoid an error when the iopub channel is
+                # closed.
+                # See jupyter/qtconsole#574
+                if not self.kernel_client.iopub_channel.closed():
+                    self.kernel_client.iopub_channel.flush()
+            except AttributeError:
+                self.kernel_client.iopub_channel.flush()
+
             self._waiting_pdb_input = True
-            self._readline(prompt=prompt, callback=self._pdb_readline_callback,
+            self._readline(prompt=prompt,
+                           callback=self.pdb_execute,
                            password=password)
             self._executing = False
             self._highlighter.highlighting_on = True
             # The previous code finished executing
             self.executed.emit(self._pdb_prompt)
             self.sig_pdb_prompt_ready.emit()
-            self.sig_pdb_state_changed.emit(True, self.get_pdb_last_step())
+            self.sig_pdb_state_changed.emit(True)
 
         self._pdb_input_ready = True
 
@@ -654,13 +660,23 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
             self.pdb_execute(*args)
             return
 
+    def pdb_input_reply(self, line, echo_stack_entry):
+        """Send a pdb input to the kernel."""
+        if self._pdb_prompt_input:
+            # Send line to input
+            self.kernel_client.input(line)
+            return
+
+        self.call_kernel(interrupt=True).pdb_input_reply(
+            line, echo_stack_entry=echo_stack_entry)
+
     # --- Private API (overrode by us) ----------------------------------------
     def _show_prompt(self, prompt=None, html=False, newline=True,
                      separator=True):
         """
         Writes a new prompt at the end of the buffer.
         """
-        if prompt == self._pdb_prompt[0]:
+        if prompt == self._pdb_prompt:
             html = True
             prompt = '<span class="in-prompt">%s</span>' % prompt
         super(DebuggingWidget, self)._show_prompt(prompt, html, newline,
@@ -687,7 +703,7 @@ class DebuggingWidget(DebuggingHistoryWidget, SpyderConfigurationAccessor):
         """Call the callback with the result of is_complete."""
         # Add a continuation prompt if not complete
         if self.is_waiting_pdb_input():
-            # As the work is done on this side, check syncronously.
+            # As the work is done on this side, check synchronously.
             complete, indent = self._is_pdb_complete(source)
             callback(complete, indent)
         else:
